/**
 * Modules registry
 * @type {Object<string,any>}
 */
const modules = {};
/**
 * @type {Object<string,any>}
 */
const instance = {};
/**
 * @param {string} name
 * @param {Object} [params]
 * @returns {{}}
 */
const newModule = function (name, params) {
  if (!modules[name]) {
    throw Error(['module is undefined', name].join(' '))
  }
  return new modules[name](params || {})
};

const addModule = function (name, module) {
  if (modules[name]) {
    throw Error(['module already added', name].join(' '))
  }
  modules[name] = module;
};

function Component(name, params) {
  if (instance[name]) return instance[name]
  return (instance[name] = newModule(name, params))
}

Component.get = function (name, params) {
  return newModule(name, params)
};

Component.add = function (name, module) {
  addModule(name, module);
  return this
};

const digitTest = /^\d+$/;
const keyBreaker = /([^\[\]]+)|(\[\])/g;
const paramTest = /([^?#]*)(#.*)?$/;
const r20 = /%20/g;
const rbracket = /\[\]$/;
const matchSymbol = /\s([a-zA-Z]+)/;

const prep = (str) => {
  return decodeURIComponent(str.replace(/\+/g, ' '))
};
const flatten = (array) => {
  let result = [];
  forEach(array, (item) => {
    result = result.concat(isArray(item) ? flatten(item) : item);
  });
  return result
};

const recursiveParams = (prefix, obj, next) => {
  if (isArray(obj)) {
    forEach(obj, (item, prop) => {
      if (rbracket.test(prefix)) {
        next(prefix, item);
      } else {
        recursiveParams(prefix + '[' + (isPlainObject(obj) ? prop : '') + ']', item, next);
      }
    });
  } else if (isPlainObject(obj)) {
    forEach(obj, (item, prop) => {
      recursiveParams(prefix + '[' + prop + ']', item, next);
    });
  } else {
    next(prefix, obj);
  }
};

const getType = (o) => {
  return {}.toString.call(o).match(matchSymbol)[1].toLowerCase()
};
const isPlainObject = (o) => {
  return !!o && typeof o === 'object' && o.constructor === Object
};

const isFunction = (o) => {
  return getType(o) === 'function'
};

const isArray = (o) => {
  return getType(o) === 'array'
};

const isElement = (o) => {
  return o && o.nodeType === 1
};

const toArray = (o) => {
  return [].slice.call(o)
};

const hasProp = (o, v) => {
  return o && o.hasOwnProperty(v)
};

const forEach = (ob, cb, cx) => {
  for (let p in ob) if (hasProp(ob, p)) cb.call(cx || null, ob[p], p);
};

const cleanObject = (ob) => {
  for (let p in ob) {
    if (hasProp(ob, p)) {
      if (ob[p].length === 0) {
        if (isArray(ob)) ob.splice(p, 1);
        if (isPlainObject(ob)) delete ob[p];
      } else if (isPlainObject(ob[p])) {
        cleanObject(ob[p]);
      }
    }
  }
  return ob
};

const extend = (obj, ...args) => {
  forEach(args, function (o) {
    if (o !== null) {
      forEach(o, function (value, key) {
        if (isPlainObject(value)) {
          obj[key] = extend(obj[key] || {}, value);
        } else {
          obj[key] = value;
        }
      });
    }
  });
  return obj
};

const uuid = () => {
  let a = 0,
    b = '';
  while (a++ < 36) {
    if ((a * 51) & 52) {
      b += (a ^ 15 ? 8 ^ (Math.random() * (a ^ 20 ? 16 : 4)) : 4).toString(16);
    } else {
      b += '-';
    }
  }
  return b
};

const isObject = (o) => {
  return getType(o) === 'object'
};

const isRegexp = (o) => {
  return getType(o) === 'regexp'
};

const isArguments = (o) => {
  return getType(o) === 'arguments'
};

const isError = (o) => {
  return getType(o) === 'error'
};

const isDate = (o) => {
  return getType(o) === 'date'
};

const isString = (o) => {
  return getType(o) === 'string'
};

const isNumber = (o) => {
  return getType(o) === 'number'
};

const map = (ob, cb, cx) => {
  let p,
    t,
    r = [];
  for (p in ob) if (hasProp(ob, p)) if ((t = cb.call(cx || null, ob[p], p)) !== undefined) r[p] = t;
  return r
};

const querySelectorAll = (o, p) => {
  return toArray((p || document).querySelectorAll(o))
};

const querySelector = (o, p) => {
  return (p || document).querySelector(o)
};

const deparam = (params) => {
  const data = {};
  if (params.charAt(0) === '?') params = params.slice(1);
  if (params && paramTest.test(params)) {
    forEach(params.split('&'), (pair) => {
      let parts = pair.split('='),
        key = prep(parts.shift()),
        value = prep(parts.join('=')),
        current = data;
      if (key) {
        parts = key.match(keyBreaker);
        for (let j = 0, l = parts.length - 1; j < l; j++) {
          if (!current[parts[j]]) {
            current[parts[j]] = digitTest.test(parts[j + 1]) || parts[j + 1] === '[]' ? [] : {};
          }
          current = current[parts[j]];
        }
        const lastPart = parts.pop();
        if (lastPart === '[]') {
          current.push(value);
        } else {
          current[lastPart] = value;
        }
      }
    });
  }
  return data
};

const param = (obj) => {
  const result = [];
  forEach(obj, (item, prop) => {
    recursiveParams(prop, item, (key, value) => {
      value = value == null ? '' : value;
      result.push([encodeURIComponent(key), encodeURIComponent(value)].join('='));
    });
  });
  return result.join('&').replace(r20, '+')
};

const removeElement = (el) => {
  el.parentNode.removeChild(el);
};

const createElement = (tag) => {
  return document.createElement(tag)
};

const addClass = (...args) => {
  const el = args.shift();
  if (isElement(el) === false) return
  const classList = el.className.trim().split(/\s+/);
  const tokens = flatten(args.map((item) => item.trim().split(/\s+/)));
  tokens.forEach((token) => {
    if (token && !~classList.indexOf(token)) {
      classList.push(token);
    }
  });
  el.className = classList.join(' ').trim();
};

const removeClass = (...args) => {
  const el = args.shift();
  if (isElement(el) === false) return
  const classList = el.className.trim().split(/\s+/);
  const tokens = flatten(args.map((item) => item.trim().split(/\s+/)));
  tokens.forEach((token) => {
    if (token) {
      const index = classList.indexOf(token);
      if (!!~index) {
        classList.splice(index, 1);
      }
    }
  });
  el.className = classList.join(' ').trim();
};

const removeAttr = (el, attrs) => {
  if (isElement(el) === false) return false
  if (isPlainObject(attrs)) {
    forEach(attrs, (value) => {
      el.removeAttribute(value);
    });
  }
};

const addAttr = (el, attrs) => {
  if (isElement(el) === false) return false
  if (isPlainObject(attrs)) {
    forEach(attrs, (value, name) => {
      el.setAttribute(name, value);
    });
  }
};

const getStyle = (el, prop, getComputedStyle) => {
  getComputedStyle = window.getComputedStyle;
  return (getComputedStyle ? getComputedStyle(el) : el['currentStyle'])[
    prop.replace(/-(\w)/gi, function (word, letter) {
      return letter.toUpperCase()
    })
  ]
};

const getPath = (path) => {
  const props = path.split('.');
  const first = props.shift();
  let value = null;
  if (hasProp(window, first)) {
    value = window[first];
    forEach(props, (name) => {
      value = hasProp(value, name) ? value[name] : null;
    });
  }
  return value
};

const stringFormat = (format, params, expr) => {
  return (format || '').replace(expr || /{(.+?)}/g, (match, prop) => {
    return params[prop] || match
  })
};

const cssUnit = (value, unit) => {
  return String(value || 0)
    .concat(unit || '')
    .concat(' !important')
};
/**
 * @function
 * @type {function(*, *, *)}
 */
const getPaymentRequest = ((cx) => {
  let NativePaymentRequest;
  if (hasProp(cx, 'PaymentRequest') && isFunction(cx.PaymentRequest)) {
    NativePaymentRequest = cx.PaymentRequest;
  }
  return (methods, details, options) => {
    let request = null;
    details = details || {};
    details.id = uuid();
    options = options || {};
    if (NativePaymentRequest) {
      try {
        request = new NativePaymentRequest(methods, details, options);
      } catch (e) {
        request = null;
      }
    }
    return request
  }
})(window);

const jsonParse = (value, defaults) => {
  try {
    return JSON.parse(value)
  } catch (e) {
    return defaults
  }
};

const PromiseWithResolvers = () => {
  let resolve;
  let reject;
  const promise = new Promise((promiseResolve, promiseReject) => {
    resolve = promiseResolve;
    reject = promiseReject;
  });
  return { promise, resolve, reject }
};

/**
 * @typedef PaymentRequestMethod
 * @property {{[string]:any}} data
 */

/**
 * @typedef PaymentRequestOptions
 * @property {boolean} [requestPayerEmail]
 * @property {boolean} [requestPayerPhone]
 * @property {boolean} [requestPayerName]
 * @property {boolean} [requestShipping]
 */

/**
 * @typedef PaymentRequestConfig
 * @property [PaymentRequestMethod] methods
 * @property {{[string]:any}} options
 */

/**
 * @param {PaymentRequestConfig} config
 */
const getAppleSessionConfig = (config) => {
  const { methods, details, options } = config;
  const method = methods.find((item) => item.supportedMethods === 'https://apple.com/apple-pay');
  const requiredShippingContactFields = [];
  const requiredBillingContactFields = [];
  if (options.requestShipping) requiredBillingContactFields.push('postalAddress');
  if (options.requestPayerName) requiredShippingContactFields.push('name');
  if (options.requestPayerEmail) requiredShippingContactFields.push('email');
  if (options.requestPayerPhone) requiredShippingContactFields.push('phone');
  //method.data.recurringPaymentRequest = recurringPaymentRequest
  //method.data.deferredPaymentRequest = deferredPaymentRequest
  return {
    version: method.data.version,
    countryCode: method.data.countryCode,
    supportedNetworks: method.data.supportedNetworks,
    merchantCapabilities: method.data.merchantCapabilities,
    merchantIdentifier: method.data.merchantIdentifier,
    lineItems: method.data.lineItems || details.displayItems,
    additionalLineItems: method.data.additionalLineItems,
    recurringPaymentRequest: method.data.recurringPaymentRequest,
    deferredPaymentRequest: method.data.deferredPaymentRequest,
    //
    currencyCode: method.data.currencyCode || details.total.amount.currency,
    total: {
      label: details.total.label,
      amount: details.total.amount.value,
    },
    //
    requiredBillingContactFields,
    requiredShippingContactFields,
  }
};

// export const deferredPaymentRequest = {
//   paymentDescription: 'Бронювання готельного номеру',
//   deferredBilling: {
//     label: 'Бронювання',
//     amount: '50.00',
//     type: 'final',
//     paymentTiming: 'deferred',
//     deferredPaymentDate: new Date('2026-01-05'),
//   },
//   managementURL: 'https://flitt.ksv.app/manage-booking.html',
// }

// export const recurringPaymentRequest = {
//   paymentDescription: 'Місячна підписка',
//   regularBilling: {
//     label: 'Підписка на місяць',
//     amount: '20.00',
//     type: 'final',
//     paymentTiming: 'recurring',
//     recurringPaymentStartDate: new Date('2026-01-01T00:00:00'),
//     recurringPaymentIntervalUnit: 'month',
//     recurringPaymentIntervalCount: 6,
//     recurringPaymentEndDate: new Date('2028-01-01T00:00:00'),
//   },
//   managementURL: 'https://flitt.ksv.app/manage-subscription.html',
// }

const getFunction = (path) => {
  const props = path.split('.');
  const value = props.reduce((acc, cur) => {
    if (acc && acc.hasOwnProperty(cur)) {
      acc = acc[cur];
      return acc
    }
    return null
  }, window);
  return isFunction(value) ? value : value
};

const loadExternalApi = (url, path) => {
  const [existScript] = toArray(document.scripts).filter((script) => script.src === url);
  if (existScript) return Promise.resolve(getFunction(path))
  const { promise, resolve } = PromiseWithResolvers();
  const script = document.createElement('script');
  script.async = true;
  script.src = url;
  script.addEventListener('load', () => {
    const value = getFunction(path);
    console.log('api loading', 'success', url);
    resolve(value);
  });
  script.addEventListener('error', (e) => {
    const value = getFunction(path);
    console.log('api loading', 'error', url);
    resolve(value);
  });
  document.head.appendChild(script);
  return promise
};

var Utils = /*#__PURE__*/Object.freeze({
  __proto__: null,
  PromiseWithResolvers: PromiseWithResolvers,
  addAttr: addAttr,
  addClass: addClass,
  cleanObject: cleanObject,
  createElement: createElement,
  cssUnit: cssUnit,
  deparam: deparam,
  extend: extend,
  forEach: forEach,
  getAppleSessionConfig: getAppleSessionConfig,
  getFunction: getFunction,
  getPath: getPath,
  getPaymentRequest: getPaymentRequest,
  getStyle: getStyle,
  getType: getType,
  hasProp: hasProp,
  isArguments: isArguments,
  isArray: isArray,
  isDate: isDate,
  isElement: isElement,
  isError: isError,
  isFunction: isFunction,
  isNumber: isNumber,
  isObject: isObject,
  isPlainObject: isPlainObject,
  isRegexp: isRegexp,
  isString: isString,
  jsonParse: jsonParse,
  loadExternalApi: loadExternalApi,
  map: map,
  param: param,
  querySelector: querySelector,
  querySelectorAll: querySelectorAll,
  removeAttr: removeAttr,
  removeClass: removeClass,
  removeElement: removeElement,
  stringFormat: stringFormat,
  toArray: toArray,
  uuid: uuid
});

const ApiOrigin = 'https://pay.flitt.com';

const ApiEndpoint = {
  gateway: '/latest/checkout-v2/index.html',
  element: '/latest/checkout-v2/button/element.html',
};

const ProxyUrl = 'http://secure-redirect.cloudipsp.com/submit/';

const GooglePayApi = 'https://pay.google.com/gp/p/js/pay.js';

const GoogleBaseRequest = {
  apiVersion: 2,
  apiVersionMinor: 0,
  allowedPaymentMethods: [
    {
      type: 'CARD',
      parameters: {
        allowedAuthMethods: ['PAN_ONLY', 'CRYPTOGRAM_3DS'],
        allowedCardNetworks: ['AMEX', 'DISCOVER', 'INTERAC', 'JCB', 'MASTERCARD', 'VISA'],
      },
    },
  ],
};

const PaymentMethodsConfig = [
  {
    url: 'https://pay.google.com/gp/p/js/pay.js',
    name: 'google',
    path: 'google.payments.api.PaymentsClient',
    callback(PaymentsClient) {
      if (PaymentsClient) {
        return new PaymentsClient({ environment: 'TEST' })
          .isReadyToPay(GoogleBaseRequest)
          .catch((e) => {
            console.log(e);
            return false
          })
      }
      return false
    },
  },
  {
    url: 'https://applepay.cdn-apple.com/jsapi/1.latest/apple-pay-sdk.js',
    name: 'apple',
    path: 'ApplePaySession',
    callback(ApplePaySession) {
      if (ApplePaySession) {
        return ApplePaySession.supportsVersion(3)
      }
      return false
    },
  },
];

const GooglePayLanguages = [
  'ar',
  'bg',
  'ca',
  'zh',
  'hr',
  'cs',
  'da',
  'nl',
  'en',
  'et',
  'fi',
  'fr',
  'de',
  'el',
  'id',
  'it',
  'ja',
  'ko',
  'ms',
  'no',
  'pl',
  'pt',
  'ru',
  'sr',
  'sk',
  'sl',
  'es',
  'sv',
  'th',
  'tr',
  'uk',
];

const PaymentRequestDetails = {
  total: {
    label: 'Total',
    amount: {
      currency: 'USD',
      value: '0.00',
    },
  },
};

const ApiFrameCss = {
  width: '1px !important',
  height: '1px !important',
  left: '1px !important',
  bottom: '1px !important',
  position: 'fixed !important',
  border: '0px !important',
};

const ButtonFrameCss = {
  border: 'none !important',
  margin: '0 !important',
  padding: '0 !important',
  display: 'block !important',
  width: '1px !important',
  'min-width': '100% !important',
  background: 'transparent !important',
  position: 'relative !important',
  opacity: '0 !important',
  overflow: 'hidden !important',
  height: '100% !important',
  outline: 'none !important',
  'z-index': '1 !important',
};

const ButtonFrameAttrs = {
  tabindex: '-1',
  scrolling: 'no',
  frameborder: 0,
  allowtransparency: true,
  allowpaymentrequest: true,
};

const ButtonCoverCss = {
  'z-index': '2 !important',
  position: 'absolute !important',
  border: 'none !important',
  background: 'transparent !important',
  left: '0 !important',
  top: '0 !important',
  cursor: 'pointer !important',
  outline: 'none !important',
  width: '100% !important',
  height: '100% !important',
};

const ButtonCoverAttrs = {
  role: 'button',
  'aria-pressed': 'false',
};

const ButtonContainerCss = {
  border: '0 !important',
  margin: '0 !important',
  padding: '0 !important',
  display: 'block !important',
  background: 'transparent !important',
  'user-select': 'none !important',
  overflow: 'hidden !important',
  position: 'relative !important',
  opacity: '1 !important',
  height: '0 !important',
  width: '100% !important',
  outline: 'none !important',
};

const ButtonDefaultColor = 'dark';

const ButtonColorMap = {
  dark: 'dark',
  light: 'light',
  black: 'dark',
  white: 'light',
};

const ButtonLabelMap = {
  ar: '',
  bg: '',
  ca: '',
  zh: '',
  hr: '',
  cs: '',
  da: '',
  nl: '',
  en: 'Pay with',
  et: '',
  es: 'Comprar con',
  el: '',
  fi: '',
  fr: 'Acheter avec',
  de: 'Zahlen über',
  id: '',
  it: 'Acquista con',
  ja: '',
  ko: '',
  ms: '',
  no: '',
  pl: 'Zapłać przez',
  pt: '',
  ru: 'Оплатить через',
  sr: '',
  sk: 'Zaplatiť cez',
  sl: '',
  sv: '',
  th: '',
  tr: '',
  uk: 'Оплатити через',
};

var Config = /*#__PURE__*/Object.freeze({
  __proto__: null,
  ApiEndpoint: ApiEndpoint,
  ApiFrameCss: ApiFrameCss,
  ApiOrigin: ApiOrigin,
  ButtonColorMap: ButtonColorMap,
  ButtonContainerCss: ButtonContainerCss,
  ButtonCoverAttrs: ButtonCoverAttrs,
  ButtonCoverCss: ButtonCoverCss,
  ButtonDefaultColor: ButtonDefaultColor,
  ButtonFrameAttrs: ButtonFrameAttrs,
  ButtonFrameCss: ButtonFrameCss,
  ButtonLabelMap: ButtonLabelMap,
  GoogleBaseRequest: GoogleBaseRequest,
  GooglePayApi: GooglePayApi,
  GooglePayLanguages: GooglePayLanguages,
  PaymentMethodsConfig: PaymentMethodsConfig,
  PaymentRequestDetails: PaymentRequestDetails,
  ProxyUrl: ProxyUrl
});

const PENDING = 0;
const RESOLVED = 1;
const REJECTED = 2;
const PROGRESS = 3;

function Deferred() {
  const resolveCallbacks = [];
  const rejectCallbacks = [];
  const notifyCallbacks = [];
  let status = PENDING;
  let resultContext = undefined;
  let resultArgs = null;
  let chainArgs = null;
  const executeCallback = (context, args, callbacks) => {
    const list = [].concat(callbacks);
    const save = (chainArgs, isArgs) => {
      if (isArgs) {
        resultArgs = Array.from(chainArgs);
      } else {
        resultArgs = [chainArgs];
      }
    };
    const execute = () => {
      const item = list.shift();
      if (item) {
        chainArgs = item.apply(resultContext, resultArgs);
        if (chainArgs !== undefined) {
          if (chainArgs && chainArgs.then) {
            chainArgs.then((...chainArgs) => {
              save(chainArgs, true);
              execute();
            });
          } else {
            save(chainArgs);
            execute();
          }
        } else {
          execute();
        }
      }
    };
    resultContext = context;
    resultArgs = args;
    execute();
  };
  /**
   * @lends Deferred.prototype
   */
  const promise = {
    done(callback) {
      if (isFunction(callback)) {
        if (this.isResolved()) {
          callback.apply(resultContext, resultArgs);
        } else {
          resolveCallbacks.push(callback);
        }
      }
      return this
    },
    fail(callback) {
      if (isFunction(callback)) {
        if (this.isRejected()) {
          callback.apply(resultContext, resultArgs);
        } else {
          rejectCallbacks.push(callback);
        }
      }
      return this
    },
    progress(callback) {
      if (isFunction(callback)) {
        if (this.isProgress()) {
          callback.apply(resultContext, resultArgs);
        } else {
          notifyCallbacks.push(callback);
        }
      }
      return this
    },
    always(callback) {
      return this.done(callback).fail(callback)
    },
    then(resolve, reject, progress) {
      if (isFunction(resolve)) {
        this.done(resolve);
      }
      if (isFunction(reject)) {
        this.fail(reject);
      }
      if (isFunction(progress)) {
        this.progress(reject);
      }
      return this
    },
    promise(object) {
      if (object === null) {
        return promise
      }
      forEach(promise, (item, prop) => {
        object[prop] = item;
      });
      return object
    },
    state: function () {
      return status
    },
    isPending() {
      return status === PENDING
    },
    isRejected() {
      return status === REJECTED
    },
    isResolved() {
      return status === RESOLVED
    },
    isProgress() {
      return status === PROGRESS
    },
    resolveWith(context, args) {
      if (this.isPending()) {
        status = RESOLVED;
        executeCallback(context, args, resolveCallbacks);
      }
      return this
    },
    rejectWith(context, args) {
      if (this.isPending()) {
        status = REJECTED;
        executeCallback(context, args, rejectCallbacks);
      }
      return this
    },
    notifyWith(context, args) {
      if (this.isPending()) {
        status = PROGRESS;
        executeCallback(context, args, notifyCallbacks);
        status = PENDING;
      }
      return this
    },
    resolve() {
      return this.resolveWith(this, Array.from(arguments))
    },
    reject() {
      return this.rejectWith(this, Array.from(arguments))
    },
    notify() {
      return this.notifyWith(this, Array.from(arguments))
    },
  };
  return promise
}

let init = false;
const fnTest = /xyz/.test(
  function () {
    return 'xyz'
  }.toString()
)
  ? /\b_super\b/
  : /.*/;

function ClassObject() {}

ClassObject.prototype._super = function () {};

ClassObject.prototype.instance = function (params) {
  return new this.constructor(params)
};

ClassObject.prototype.proxy = function (fn) {
  fn = typeof fn == 'string' ? this[fn] : fn;
  return (function (cx, cb) {
    return function () {
      return cb.apply(cx, [this].concat(Array.prototype.slice.call(arguments)))
    }
  })(this, fn)
};

function superMethod(parent, name, method) {
  return function () {
    let temp = this._super,
      result;
    this._super = parent[name];
    result = method.apply(this, arguments);
    this._super = temp;
    return result
  }
}

function assign(target, instance) {
  let prop,
    proto,
    parent = target.prototype;
  init = true;
  proto = new target();
  init = false;
  for (prop in instance) {
    if (instance.hasOwnProperty(prop)) {
      if (
        typeof parent[prop] == 'function' &&
        typeof instance[prop] == 'function' &&
        fnTest.test(instance[prop])
      ) {
        proto[prop] = superMethod(parent, prop, instance[prop]);
      } else {
        proto[prop] = instance[prop];
      }
    }
  }
  return proto
}

ClassObject.extend = function extend(instance) {
  function Class() {
    if (!init && this) this['init'].apply(this, arguments);
  }
  Class.prototype = assign(this, instance);
  Class.prototype.constructor = Class;
  Class.extend = extend;
  return Class
};

const Event = ClassObject.extend({
  init() {
    this.events = {};
    this.empty = [];
  },
  on(type, callback) {
(this.events[type] = this.events[type] || []).push(callback);
    return this
  },
  off(type, callback) {
    type || (this.events = {});
    let list = this.events[type] || this.empty,
      i = (list.length = callback ? list.length : 0);
    while (i--) callback === list[i] && list.splice(i, 1);
    return this
  },
  trigger(type) {
    let e = this.events[type] || this.empty,
      list = e.length > 0 ? e.slice(0, e.length) : e,
      i = 0,
      j;
    while ((j = list[i++])) j.apply(j, this.empty.slice.call(arguments, 1));
    return this
  },
});

const Module = ClassObject.extend({
  utils: Utils,
  getListener() {
    if (!this._listener_) this._listener_ = new Event();
    return this._listener_
  },
  destroy() {
    this.off();
  },
  on(type, callback) {
    this.getListener().on(type, callback);
    return this
  },
  off(type, callback) {
    this.getListener().off(type, callback);
    return this
  },
  proxy(fn) {
    if (!this._proxy_cache_) this._proxy_cache_ = {};
    if (this.utils.isString(fn)) {
      if (!this._proxy_cache_[fn]) {
        this._proxy_cache_[fn] = this._super(fn);
      }
      return this._proxy_cache_[fn]
    }
    return this._super(fn)
  },
  trigger() {
    this.getListener().trigger.apply(this.getListener(), arguments);
    return this
  },
  each(ob, cb) {
    this.utils.forEach(ob, this.proxy(cb));
  },
  addAttr(el, ob) {
    if (!this.utils.isElement(el)) return false
    this.utils.forEach(ob, function (v, k) {
      el.setAttribute(k, v);
    });
  },
  addCss(el, ob) {
    if (!this.utils.isElement(el)) return false
    this.utils.forEach(
      ob,
      function (v, k) {
        this.addCssProperty(el, k, v);
      },
      this
    );
  },
  addCssProperty(el, style, value) {
    let result = el.style.cssText.match(
        new RegExp('(?:[;\\s]|^)(' + style.replace('-', '\\-') + '\\s*:(.*?)(;|$))')
      ),
      idx;
    if (result) {
      idx = result.index + result[0].indexOf(result[1]);
      el.style.cssText =
        el.style.cssText.substring(0, idx) +
        style +
        ': ' +
        value +
        ';' +
        el.style.cssText.substring(idx + result[1].length);
    } else {
      el.style.cssText += ' ' + style + ': ' + value + ';';
    }
  },
  addEvent(el, ev, cb) {
    if (!el || !ev || !cb) return false
    cb = this.proxy(cb);
    if (el.addEventListener) el.addEventListener(ev, cb);
    else if (el['attachEvent']) el['attachEvent']('on' + ev, cb);
  },
  removeEvent(el, ev, cb) {
    if (!el || !ev || !cb) return false
    cb = this.proxy(cb);
    if (el.removeEventListener) el.removeEventListener(ev, cb, false);
    else if (el['detachEvent']) el['detachEvent']('on' + ev, cb);
  },
  prepareForm(url, data, target, method) {
    const form = this.utils.createElement('form');
    this.addAttr(form, {
      action: url,
      target: target || '_self',
      method: method || 'POST',
    });
    this.addCss(form, {
      display: 'none',
    });
    this.utils.forEach(
      data,
      function (v, k, el) {
        el = this.utils.createElement('input');
        el.type = 'hidden';
        el.name = k;
        el.value = v;
        form.appendChild(el);
      },
      this
    );
    return form
  },
});

const Connector = Module.extend({
  ns: 'crossDomain',
  origin: '*',
  uniqueId: 1,
  signature: null,
  init(params) {
    this.setTarget(params.target);
    this.setOrigin(params.origin);
    this.create();
  },
  create() {
    this.addEvent(window, 'message', 'router');
  },
  setOrigin(origin) {
    this.origin = origin || '*';
  },
  setTarget(target) {
    this.target = target;
    return this
  },
  getUID() {
    return ++this.uniqueId
  },
  destroy() {
    this.removeEvent(window, 'message', 'router');
    this._super();
  },
  router(window, ev, response) {
    if (this.target !== ev.source) return false
    try {
      response = JSON.parse(ev.data);
    } catch (e) {}
    if (response && response.action && response.data) {
      this.trigger(response.action, response.data);
    }
  },
  send(action, data, request, options) {
    if (!this.target) {
      return
    }
    request = JSON.stringify({
      action: action,
      data: data,
    });
    options = {
      targetOrigin: this.origin,
      delegate: 'payment',
    };
    try {
      this.target.postMessage(request, options);
    } catch (e) {
      this.target.postMessage(request, this.origin, []);
    }
  },
});

var styles = (data) => `
<style>
    .flitt-modal {
        box-sizing: border-box;
        margin: 100px auto;
        max-width: 680px;
        background-color: #fff;
        border-radius: 5px;
        box-shadow: 0 2px 2px rgba(0, 0, 0, 0.2);
        overflow: hidden;
    }

    @media (max-width: 850px) {
        .flitt-modal {
            margin: 50px auto;
        }
    }

    @media (max-width: 695px) {
        .flitt-modal {
            max-width: 100%;
            margin: 5px;
        }
    }

    .flitt-modal-wrapper {
        box-sizing: border-box;
        overflow: auto;
        position: fixed;
        z-index: 99999;
        left: 0;
        bottom: 0;
        top: 0;
        right: 0;
        background-color: rgba(0, 0, 0, 0.2);
    }

    .flitt-modal-header {
        box-sizing: border-box;
        background-color: #fafafa;
        height: 50px;
        box-shadow: 0 0 2px rgba(0, 0, 0, 0.2);
        border-top-left-radius: 5px;
        border-top-right-radius: 5px;
    }

    .flitt-modal-close {
        box-sizing: border-box;
        float: right;
        overflow: hidden;
        height: 50px;
        text-decoration: none;
        border-top-right-radius: 5px;
        color: #949494;
    }

    .flitt-modal-close:hover, .flitt-modal-close:focus, .flitt-modal-close:active {
        text-decoration: none;
        color: #646464;
    }

    .flitt-modal-close:before {
        content: "×";
        font-size: 50px;
        line-height: 50px;
        padding: 0 10px;
    }

    .flitt-modal-title {
        box-sizing: border-box;
        border-top-left-radius: 5px;
        line-height: 20px;
        height: 50px;
        padding: 5px 15px;
        font-size: 12px;
        display: table-cell;
        vertical-align: middle;
    }
    .flitt-modal-content {
        box-sizing: border-box;
        border-bottom-left-radius: 5px;
        min-height: 650px;
    }
    .flitt-modal-iframe {
        overflow-x: hidden;
        border: 0;
        display: block;
        width: 100%;
        height: 750px;
    }
</style>
`;

var acs = (data) => `
${styles()}
<div class="flitt-modal-wrapper">
    <div class="flitt-modal">
        <div class="flitt-modal-header">
            <a href="javascript:void(0)" class="flitt-modal-close"></a>
            <div class="flitt-modal-title">
                ${data.messages.modalHeader}
                <a href='javascript:void(0)'>${data.messages.modalLinkLabel}</a>
            </div>
        </div>
        <div class="flitt-modal-content">
            <iframe src="about:blank" class="flitt-modal-iframe" frameborder="0" allowtransparency="true"></iframe>
        </div>
    </div>
</div>
`;

var Views = /*#__PURE__*/Object.freeze({
  __proto__: null,
  acs: acs
});

function empty(name) {
  return function (data) {
    return ['template', name, 'not found'].join(' ')
  }
}

const Template = ClassObject.extend({
  init(name) {
    this.view = Views[name] || empty(name);
  },
  render(data) {
    return this.view(data)
  },
});

const Modal = Module.extend({
  init(data) {
    this.checkout = data.checkout;
    this.model = data.model || {};
    this.messages = data.checkout.params.messages || {};
    this.template = new Template('acs');
    this.body = this.utils.querySelector('body');
    this.initModal();
    this.initConnector();
  },
  initModal() {
    this.name = ['modal-iframe', this.getRandomNumber()].join('-');
    this.modal = this.utils.createElement('div');
    this.modal.innerHTML = this.template.render({
      model: this.model,
      messages: this.messages,
    });
    this.iframe = this.find('.flitt-modal-iframe');
    this.addAttr(this.iframe, { name: this.name, id: this.name });
    if (this.model['send_data']) {
      this.form = this.prepareForm(this.model['url'], this.model['send_data'], this.name);
      this.modal.appendChild(this.form);
    } else {
      this.iframe.src = this.model['url'];
    }
    this.addEvent(this.find('.flitt-modal-close'), 'click', 'closeModal');
    this.addEvent(this.find('.flitt-modal-title a'), 'click', 'submitForm');
    this.initScrollbar();
    this.body.appendChild(this.modal);
    if (this.form) {
      this.form.submit();
    }
  },
  measureScrollbar() {
    const scrollDiv = document.createElement('div');
    scrollDiv.className = 'modal-scrollbar-measure';
    this.body.appendChild(scrollDiv);
    this.utils.removeElement(scrollDiv);
    return scrollDiv.offsetWidth - scrollDiv.clientWidth
  },
  checkScrollbar() {
    let documentElementRect;
    let fullWindowWidth = window.innerWidth;
    if (!fullWindowWidth) {
      documentElementRect = document.documentElement.getBoundingClientRect();
      fullWindowWidth = documentElementRect.right - Math.abs(documentElementRect.left);
    }
    this.bodyIsOverflowing = document.body.clientWidth < fullWindowWidth;
    this.scrollbarWidth = this.measureScrollbar();
  },
  initScrollbar() {
    this.checkScrollbar();
    this.bodyPad = parseInt(this.utils.getStyle(this.body, 'padding-right') || 0, 10);
    this.originalBodyPad = document.body.style.paddingRight || '';
    this.originalOverflow = document.body.style.overflow || '';
    if (this.bodyIsOverflowing) {
      this.addCss(this.body, {
        paddingRight: [this.bodyPad + this.scrollbarWidth, 'px'].join(''),
        overflow: 'hidden',
      });
    }
  },
  resetScrollbar() {
    this.addCss(this.body, {
      paddingRight: this.originalBodyPad ? [this.originalBodyPad, 'px'].join('') : '',
      overflow: this.originalOverflow,
    });
  },
  getRandomNumber() {
    return Math.round(Math.random() * 1000000000)
  },
  find(selector) {
    return this.utils.querySelector(selector, this.modal)
  },
  closeModal(el, ev) {
    ev.preventDefault();
    this.trigger('close', this.data);
    this.removeModal();
  },
  submitForm(el, ev) {
    ev.preventDefault();
    this.trigger('submit', this.data);
    this.addAttr(this.form, {
      target: '_blank',
    });
    this.form.submit();
  },
  removeModal() {
    this.destroy();
  },
  destroy() {
    this.utils.removeElement(this.modal);
    this.resetScrollbar();
    this.connector.destroy();
    this._super();
  },
  initConnector() {
    this.connector = new Connector({ target: this.iframe.contentWindow });
    this.connector.on('response', this.proxy('onResponse'));
  },
  onResponse(ev, data) {
    this.sendResponse(data);
    this.removeModal();
  },
  sendResponse(data) {
    this.checkout.connector.send('request', {
      uid: data.uid,
      action: 'api.checkout.proxy',
      method: 'send',
      params: data,
    });
  },
});

const Model = Module.extend({
  init(data) {
    this.data = data || {};
    this.create();
  },
  create() {},
  eachProps(args) {
    const name = args[1] ? args[0] : null;
    const callback = args[1] ? args[1] : args[0];
    const list = name ? this.alt(name, []) : this.data;
    return {
      list: list,
      callback: callback,
    }
  },
  each() {
    const props = this.eachProps(arguments);
    for (let prop in props.list) {
      if (props.list.hasOwnProperty(prop)) {
        props.callback(this.instance(props.list[prop]), props.list[prop], prop);
      }
    }
  },
  filter() {
    const props = this.eachProps(arguments);
    for (let prop in props.list) {
      if (props.list.hasOwnProperty(prop)) {
        const item = this.instance(props.list[prop]);
        if (props.callback(item, props.list[prop], prop)) {
          return props.list[prop]
        }
      }
    }
  },
  find() {
    const props = this.eachProps(arguments);
    for (let prop in props.list) {
      if (props.list.hasOwnProperty(prop)) {
        const item = this.instance(props.list[prop]);
        if (props.callback(item, props.list[prop], prop)) {
          return item
        }
      }
    }
    return false
  },
  alt(prop, defaults) {
    prop = this.attr(prop);
    return typeof prop === 'undefined' ? defaults : prop
  },
  attr(key, value) {
    let i = 0,
      data = this.data,
      name = (key || '').split('.'),
      prop = name.pop(),
      len = arguments.length;
    for (; i < name.length; i++) {
      if (data && data.hasOwnProperty(name[i])) {
        data = data[name[i]];
      } else {
        if (len === 2) {
          data = data[name[i]] = {};
        } else {
          break
        }
      }
    }
    if (len === 1) {
      return data ? data[prop] : undefined
    }
    if (len === 2) {
      data[prop] = value;
    }
    return this
  },
  stringify() {
    return JSON.stringify(this.serialize())
  },
  serialize() {
    return this.data
  },
});

const Response = Model.extend({
  stringFormat(string) {
    const that = this;
    return (string || '').replace(/{(.+?)}/g, function (match, prop) {
      return that.attr(['order.order_data', prop].join('.')) || match
    })
  },
  setConnector(connector) {
    this.connector = connector;
    return this
  },
  setUID(uid) {
    this.uid = uid;
    return this
  },
  getUID() {
    return this.uid
  },
  formDataProxy(url, data, target, method) {
    location.assign(
      [
        ProxyUrl,
        JSON.stringify({
          url: url,
          params: data,
          target: target,
          method: method,
        }),
      ].join('#')
    );
  },
  formDataSubmit(url, data, target, method) {
    if (url.match(/^http:/)) {
      return this.formDataProxy(url, data, target, method)
    }
    const action = this.stringFormat(url);
    const form = this.prepareForm(action, data, target, method);
    const body = this.utils.querySelector('body');
    body.appendChild(form);
    form.submit();
    form.parentNode.removeChild(form);
  },
  inProgress() {
    return this.attr('order.in_progress')
  },
  readyToSubmit() {
    return this.attr('order.ready_to_submit')
  },
  waitForResponse() {
    return this.attr('order.pending')
  },
  needVerifyCode() {
    return this.attr('order.need_verify_code')
  },
  redirectUrl() {
    if (this.attr('url')) {
      this.redirectToUrl(this.attr('url'));
      return true
    }
    return false
  },
  redirectToUrl(url) {
    location.assign(url);
  },
  submitToMerchant() {
    const ready = this.attr('order.ready_to_submit');
    const url = this.attr('model.url') || this.attr('order.response_url');
    const method = this.attr('order.method');
    const action = this.attr('order.action');
    const data = this.attr('model.send_data') || this.attr('order.order_data');
    if (ready && url && data) {
      if (action === 'redirect' || data['get_without_parameters'] === true) {
        this.redirectToUrl(url);
      } else {
        this.formDataSubmit(url, data, '_self', method);
      }
      return true
    }
  },
  submitForm() {
    const method = this.attr('method');
    const url = this.attr('url');
    const data = this.attr('send_data');
    if (url && data) {
      this.formDataSubmit(url, data, '_self', method);
      return true
    }
    return false
  },
  sendResponse() {
    const action = this.attr('action');
    if (action === 'submit') return this.submitForm()
    if (action === 'redirect') return this.redirectUrl()
    return false
  },
  prepare3dsData() {
    const params = {};
    const data = this.attr('submit3ds');
    if (data['3ds']) {
      params.token = this.attr('token');
      params.uid = this.getUID();
      params.frame = true;
      if (data['send_data'].TermUrl) {
        data['send_data'].TermUrl = [data['send_data'].TermUrl, this.utils.param(params)].join(
          '#!!'
        );
      }
    }
    return data
  },
  waitOn3dsDecline() {
    const data = this.alt('submit3ds.checkout_data', {
      js_wait_on_3ds_decline: false,
      js_wait_on_3ds_decline_duration: 0,
    });
    return data.js_wait_on_3ds_decline ? data.js_wait_on_3ds_decline_duration : 0
  },
  submit3dsForm() {
    if (this.attr('submit3ds.checkout_data')) {
      this.connector.trigger('modal', this.prepare3dsData());
    }
  },
  supportedMethod(method) {
    const item = this.find('methods', function (item) {
      return item.alt('supportedMethods', '').match(method)
    });
    if (item) {
      this.attr('methods', [item.serialize()]);
    } else {
      this.attr('methods', []);
    }
  },
});

const Api = Module.extend({
  defaults: {
    origin: ApiOrigin,
    endpoint: ApiEndpoint,
    container: 'body',
    messages: {
      modalHeader:
        'Now you will be redirected to your bank 3DSecure. If you are not redirected please refer',
      modalLinkLabel: 'link',
    },
  },
  init(params) {
    this.initParams(params);
  },
  url(type, url) {
    return [this.params.origin, this.params.endpoint[type] || '/', url || ''].join('')
  },
  extendParams(params) {
    this.utils.extend(this.params, params);
    return this
  },
  initParams(params) {
    this.iframeDeferred = Deferred();
    this.params = this.utils.extend({}, this.defaults);
    this.extendParams(params);
    this.setOrigin(this.params.origin);
  },
  setOrigin(origin) {
    if (this.utils.isString(origin)) {
      this.params.origin = origin;
    }
    return this
  },
  scope(callback) {
    callback = this.proxy(callback);
    this.domReady(function () {
      this.createFrame();
      if (this.iframeDeferred.isPending()) {
        this.iframeDeferred.done(callback);
      } else {
        callback();
      }
    });
  },
  domReady(callback) {
    callback = this.proxy(callback);
    if (document.readyState === 'loading') {
      window.addEventListener('DOMContentLoaded', callback);
    } else {
      callback();
    }
  },
  request(model, method, params) {
    const defer = Deferred();
    const data = {
      uid: this.connector.getUID(),
      action: model,
      method: method,
      params: params || {},
    };
    this.connector.send('request', data);
    this.connector.on(
      data.uid,
      this.proxy(function (ev, response, model, action) {
        const responseModel = new Response(response);
        responseModel.setUID(data.uid);
        responseModel.setConnector(this.connector);
        action = 'resolveWith';
        if (responseModel.attr('submit3ds')) {
          action = 'notifyWith';
        }
        if (responseModel.attr('error')) {
          action = 'rejectWith';
        }
        defer[action](this, [responseModel]);
      })
    );
    return defer
  },
  createFrame() {
    if (this.iframe) return this
    this.iframe = this.utils.createElement('iframe');
    this.addAttr(this.iframe, {
      allowtransparency: true,
      frameborder: 0,
      scrolling: 'no',
    });
    this.addAttr(this.iframe, { src: this.url('gateway') });
    this.addCss(this.iframe, ApiFrameCss);
    if (this.utils.isElement(this.params.container)) {
      this.container = this.params.container;
    }
    if (this.utils.isString(this.params.container)) {
      this.container = this.utils.querySelector(this.params.container);
    }
    if (this.container) {
      if (this.container.firstChild) {
        this.container.insertBefore(this.iframe, this.container.firstChild);
      } else {
        this.container.appendChild(this.iframe);
      }
      this.connector = new Connector({
        target: this.iframe.contentWindow,
        origin: this.params.origin,
      });
      this.connector.on('load', this.proxy('onLoadConnector'));
      this.connector.on('modal', this.proxy('onOpenModal'));
    } else {
      throw Error(`container element not found: document.querySelector("${this.params.container}")`)
    }
    return this
  },
  onOpenModal(xhr, model) {
    this.modal = new Modal({ checkout: this, model: model });
    this.modal.on('close', this.proxy('onCloseModal'));
  },
  onCloseModal(modal, data) {
    this.trigger('modal.close', modal, data);
  },
  onLoadConnector() {
    this.iframeDeferred.resolve(true);
    this.connector.off('load');
  },
});

let supportedMethodsDeferred = null;

const getSupportedMethods = () => {
  if (supportedMethodsDeferred) return supportedMethodsDeferred
  supportedMethodsDeferred = Deferred();
  const promises = [];
  const supported = { provider: [] };
  forEach(PaymentMethodsConfig, (method) => {
    promises.push(
      loadExternalApi(method.url, method.path)
        .then(method.callback)
        .then((state) => {
          if (state) {
            supported.provider.push(method.name);
          } else {
            console.log('payment method is not supported', method.name);
          }
        })
    );
  });
  Promise.all(promises).finally(() => {
    supportedMethodsDeferred.resolve(supported);
  });
  return supportedMethodsDeferred
};

const PaymentRequestApi = Module.extend({
  config: {
    payment_system: '',
    fallback: false,
    methods: [],
    details: {},
    options: {},
  },
  supported: {
    fallback: false,
    provider: [],
  },
  payload: {
    payment_system: null,
    provider: [],
  },
  params: {},
  getSupportedMethods() {
    return getSupportedMethods().then((supported) => {
      this.setSupported(supported);
      return supported
    })
  },
  init(params) {
    this.params = params || {};
  },
  setSupported(supported) {
    this.supported = supported;
    this.trigger('supported', supported);
    return this
  },
  setPayload(payload) {
    this.payload = payload;
    this.trigger('payload', payload);
    return this
  },
  setMerchant(merchant) {
    this.merchant = merchant;
  },
  setBeforeCallback(beforeCallback) {
    if (isFunction(beforeCallback)) {
      this.params.before = beforeCallback;
    }
  },
  setAfterCallback(afterCallback) {
    if (isFunction(afterCallback)) {
      this.params.after = afterCallback;
    }
  },
  setApi(api) {
    if (api instanceof Api) this.api = api;
    return this
  },
  getProviderPayload(method) {
    return this.payload.provider[method] || {}
  },
  isMethodSupported(method) {
    return this.supported.provider.indexOf(method) !== -1
  },
  isFallbackMethod(method) {
    return method === 'google' && this.supported.fallback
  },
  request(model, method, params, success, failure) {
    const context = this;
    if (context.api) {
      context.api.scope(function () {
        context.api.request(model, method, params).done(success).fail(failure);
      });
    }
  },
  session(params) {
    const defer = Deferred();
    this.request(
      'api.checkout.pay',
      'session',
      params,
      function (model) {
        defer.resolve(model.serialize());
      },
      function (model) {
        defer.reject(model);
      }
    );
    return defer
  },
  update(data) {
    if (this.isPending()) return
    this.setPending(true);
    const context = this;
    const defer = Deferred();
    this.params.data = data;
    this.request(
      'api.checkout.pay',
      'methods',
      this.params.data,
      function (model) {
        context.setPending(false);
        context.setPayload(model.serialize());
        defer.resolve(model);
      },
      function (model) {
        context.setPending(false);
        context.trigger('error', model);
        defer.reject(model);
      }
    );
    return defer
  },
  isPending() {
    return this.pendingState === true
  },
  setPending(state) {
    this.pendingState = state;
    clearTimeout(this.pendingTimeoutEvent);
    this.pendingTimeoutEvent = setTimeout(
      function (context) {
        context.trigger('pending', state);
      },
      50,
      this
    );
  },
  before() {
    if (this.isPending()) return
    this.setPending(true);
    const defer = Deferred();
    const context = this;
    defer.always(function () {
      context.setPending(false);
    });
    this.params.before(defer);
    return defer
  },
  after(params) {
    const defer = Deferred();
    this.params.after(defer, params);
    return defer
  },
  withApplePay(response, config) {
    const sessionConfig = getAppleSessionConfig(config);
    const request = new ApplePaySession(sessionConfig.version, sessionConfig);
    const merchant = this.merchant;
    const context = this;
    request.onvalidatemerchant = function (event) {
      context
        .session({
          url: event.validationURL,
          domain: location.host,
          merchant_id: merchant,
        })
        .done((session) => {
          try {
            request.completeMerchantValidation(session.data);
          } catch (error) {
            response.reject({ code: error.code, message: error.message });
          }
        });
    };
    request.oncancel = function () {
      response.reject({ code: 20, message: 'cancel' });
    };
    request.onpaymentauthorized = function (event) {
      request.completePayment(ApplePaySession.STATUS_SUCCESS);
      response.resolve(event.payment);
    };
    request.begin();
  },
  withGooglePay(response, config) {
    const { methods } = config;
    const method = methods.find((item) => item.supportedMethods === 'https://google.com/pay');
    const client = new google.payments.api.PaymentsClient({ environment: method.data.environment });
    client
      .loadPaymentData(method.data)
      .then((paymentData) => {
        response.resolve(paymentData);
      })
      .catch((error) => {
        response.reject({ code: error.code, message: error.message });
      });
  },
  pay(method) {
    if (this.isPending()) return
    this.setPending(true);
    const context = this;
    const response = Deferred();
    const payload = this.getProviderPayload(method);
    response.always(function () {
      context.setPending(false);
    });
    switch (method) {
      case 'google':
        this.withGooglePay(response, payload);
        break
      case 'apple':
        this.withApplePay(response, payload);
        break
    }
    return response
      .done(function (data) {
        context.trigger('details', {
          payment_system: context.payload.payment_system,
          data,
        });
      })
      .fail(function (error) {
        context.trigger('error', error);
      })
  },
});

const PaymentElement = Module.extend({
  defaults: {
    method: null,
    appendTo: null,
    className: 'payment-element',
    origin: ApiOrigin,
    endpoint: ApiEndpoint.element,
    mode: 'plain',
    color: 'black',
    lang: 'en',
    height: 38,
  },
  init(params) {
    this.state = {
      isSupported: false,
      isAllowed: false,
      isMounted: false,
    };
    this.params = {};
    this.utils.extend(this.params, this.defaults);
    this.utils.extend(this.params, params);
    this.initElement();
  },
  timeout(callback, timeout) {
    this.timeoutMap = this.timeoutMap || {};
    clearTimeout(this.timeoutMap[timeout]);
    this.timeoutMap[callback] = setTimeout(callback, timeout);
    return this
  },
  getElementUrl() {
    return [this.params.origin, this.params.endpoint].join('')
  },
  getElementOptions() {
    return this.utils.param({
      method: this.params.method,
      mode: this.params.mode,
      style: this.params.style,
      color: this.params.color,
      lang: this.params.lang,
    })
  },
  initElement() {
    this.element = this.utils.createElement('div');
    this.iframe = this.utils.createElement('iframe');
    this.button = this.utils.createElement('a');
    this.addCss(this.element, ButtonContainerCss);
    this.addCss(this.button, ButtonCoverCss);
    this.addAttr(this.button, ButtonCoverAttrs);
    this.addCss(this.iframe, ButtonFrameCss);
    this.addAttr(this.iframe, ButtonFrameAttrs);
    this.addAttr(this.iframe, {
      src: [this.getElementUrl(), this.getElementOptions()].join('?'),
    });
    this.addEvent(this.iframe, 'load', 'onloadConnector');
    this.addEvent(this.iframe, 'error', 'errorConnector');
    this.addAttr(this.element, {
      class: this.params.className,
    });
    this.element.appendChild(this.iframe);
    this.element.appendChild(this.button);
  },
  errorConnector() {},
  onloadConnector() {
    this.connector = new Connector({
      target: this.iframe.contentWindow,
      origin: this.params.origin,
    });
  },
  send(action, data) {
    if (this.connector) {
      this.connector.send(action, data);
    }
  },
  onEvent(cx, ev) {
    ev.preventDefault();
    if (this.pending) return false
    this.send('event', { type: ev.type });
  },
  onClick() {
    if (this.pending) return false
    this.request.before().done(this.proxy('onClickDone')).fail(this.proxy('onClickFail'));
  },
  onClickDone() {
    this.request.pay(this.params.method);
  },
  onClickFail() {},
  onSupported(cx, supported) {
    this.state.isSupported = supported.provider.includes(this.params.method);
    this.render();
  },
  onPayload(cx, payload) {
    this.state.isAllowed = payload.allowed.includes(this.params.method);
    this.render();
  },
  onPending(cx, state) {
    this.pending = state;
    this.addCss(this.element, {
      transition: 'height 0.2s ease-out, opacity 0.4s ease-out',
      pointerEvents: this.pending ? 'none' : '',
      opacity: this.pending ? '0.5' : '1',
    });
  },
  initEvents() {
    this.addEvent(this.button, 'mouseenter', 'onEvent');
    this.addEvent(this.button, 'mouseleave', 'onEvent');
    this.addEvent(this.button, 'blur', 'onEvent');
    this.addEvent(this.button, 'focus', 'onEvent');
    this.addEvent(this.button, 'click', 'onClick');
  },
  setPaymentRequest(request) {
    if (!(request instanceof PaymentRequestApi))
      throw Error('request is not instance of PaymentRequestApi')
    this.request = request;
    const onSupported = this.proxy('onSupported');
    const onPayload = this.proxy('onPayload');
    const onPending = this.proxy('onPending');
    this.request.off('supported', onSupported).on('supported', onSupported);
    this.request.off('payload', onPayload).on('payload', onPayload);
    this.request.off('pending', onPending).on('pending', onPending);
    return this
  },
  appendTo(appendTo) {
    const container = this.utils.querySelector(appendTo);
    if (container) container.appendChild(this.element);
    this.initEvents();
    return this
  },
  render() {
    if (this.state.isSupported === false) return this
    if (this.state.isAllowed === false) return this
    this.mount();
    this.show();
  },
  unmount() {
    if (this.element.parentNode) {
      this.element.parentNode.removeChild(this.element);
    }
    return this
  },
  mount() {
    if (document.body.contains(this.element) === false) {
      this.appendTo(this.params.appendTo);
    }
    return this
  },
  show() {
    this.timeout(() => {
      this.addCss(this.iframe, {
        transition: 'opacity 0.6s 0.4s ease-out',
        opacity: this.utils.cssUnit(1),
      });
      this.addCss(this.element, {
        transition: 'height 0.2s ease-out',
        height: this.utils.cssUnit(this.params.height, 'px'),
      });
      this.trigger('show', {});
    }, 25);
    return this
  },
  hide() {
    this.timeout(() => {
      this.addCss(this.iframe, {
        transition: 'opacity 0.4s ease-out',
        opacity: this.utils.cssUnit(0),
      });
      this.addCss(this.element, {
        transition: 'height 0.2s 0.4s ease-out',
        height: this.utils.cssUnit(0, 'px'),
      });
      this.trigger('hide', {});
    }, 25);
    return this
  },
});

const PaymentButton = Module.extend({
  defaults: {
    origin: ApiOrigin,
    endpoint: ApiEndpoint,
    methods: ['apple', 'google'],
    element: null,
    style: {
      height: 38,
      mode: 'default',
      type: 'long',
      color: 'black',
    },
    data: {
      lang: 'en',
    },
    before(defer) {
      defer.resolve();
    },
    after(defer, params) {
      defer.resolve(params);
    },
  },
  init(params) {
    this.elements = {};
    this.supported = false;
    this.params = {
      element: params.element,
      methods: params.methods || this.defaults.methods,
      origin: params.origin || this.defaults.origin,
      endpoint: this.utils.extend({}, this.defaults.endpoint, params.endpoint),
      style: this.utils.extend({}, this.defaults.style, params.style),
      before: isFunction(params.before) ? params.before : this.defaults.before,
      after: isFunction(params.after) ? params.after : this.defaults.after,
      data: this.utils.extend({}, this.defaults.data, params.data),
    };
    this.initApi(params.api);
    this.initPaymentRequestApi();
    this.initElements();
    this.initData();
  },
  initData() {
    const data = this.params.data;
    if (!data.merchant_id) return
    if ((!data.currency || !data.amount) && !data.token) return
    this.update();
  },
  initApi(api) {
    if (api instanceof Api) {
      this.api = api;
    } else {
      this.api = new Api({
        origin: this.params.origin,
        endpoint: this.params.endpoint,
      });
    }
  },
  initPaymentRequestApi() {
    this.request = new PaymentRequestApi({
      before: this.params.before,
      after: this.params.after,
    });
    this.request.setApi(this.api);
    this.request.setMerchant(this.params.data.merchant_id);
    this.request.on('details', this.proxy('onDetails'));
    this.request.on('error', this.proxy('onError'));
  },
  initElements() {
    const style = this.params.style;
    const data = this.params.data;
    const origin = this.params.origin;
    const appendTo = this.params.element;
    const endpoint = this.params.endpoint.element;
    const request = this.request;
    this.container = this.utils.querySelector(this.params.element);
    this.addCss(this.container, {
      display: 'flex',
      gap: '1rem',
      'flex-direction': 'column',
    });
    forEach(this.params.methods, function (method) {
      const element = new PaymentElement({
        origin: origin,
        endpoint: endpoint,
        method: method,
        appendTo: appendTo,
        color: style.color,
        mode: style.mode,
        lang: data.lang,
        height: style.height,
      });
      element.setPaymentRequest(request);
    });
    request.getSupportedMethods();
  },
  update(data) {
    return this.request.update(this.utils.extend(this.params.data, data || {}))
  },
  onDetails(cx, data) {
    this.api.scope(() => {
      this.request.after(this.params.data).done((extendParams) => {
        this.api
          .request(
            'api.checkout.form',
            'request',
            this.utils.extend({}, this.params.data, extendParams || {}, data)
          )
          .done(this.proxy('onSuccess'))
          .fail(this.proxy('onError'));
      });
    });
  },
  onSuccess(cx, data) {
    this.trigger('success', data);
  },
  onError(cx, data) {
    this.trigger('error', data);
  },
});

const parseConfig = (config) => {
  const data = {
    payment_system: config.payment_system,
    provider: {},
    allowed: [],
  };
  const methods = config.methods || [];
  const details = config.details || {};
  const options = config.options || {};
  const regex = new RegExp('//([a-z]+)\\.com/');
  methods.forEach(function (item) {
    const match = regex.exec(item.supportedMethods);
    if (match === null) return
    data.provider[match[1]] = {
      methods: [item],
      details: details,
      options: options,
    };
  });
  return data
};

const PaymentContainer = Module.extend({
  defaults: {
    element: null,
    method: 'card',
    data: {
      lang: 'en',
    },
    style: {},
  },
  init(params) {
    this.initParams(params);
    this.initEvents();
  },
  initParams(params) {
    this.params = this.utils.extend({}, this.defaults, params);
    this.element = this.utils.querySelector(this.params.element);
    this.connector = new Connector({
      target: window.parent,
    });
    this.payment = new PaymentRequestApi({
      embedded: true,
    });
  },
  extendParams(params) {
    this.utils.extend(this.params, {
      method: params.method,
      style: params.style,
      data: params.data,
      css: params.css,
    });
  },
  getGoogleLangSupport(lang, defaults) {
    return GooglePayLanguages.indexOf(lang) !== -1 ? lang : defaults
  },
  getButtonColor(color) {
    return ButtonColorMap[color] || ButtonDefaultColor
  },
  getGoogleSvg(color, lang, mode) {
    const params = {
      endpoint: 'https://www.gstatic.com/instantbuy/svg',
      color: this.getButtonColor(color),
      mode: mode || 'plain',
      lang: lang || 'en',
    };
    let format = 'url("{endpoint}/{color}/{mode}/{lang}.svg")';
    if (mode === 'plain') {
      format = 'url("{endpoint}/{color}_gpay.svg")';
    }
    if (mode === 'buy') {
      format = 'url("{endpoint}/{color}/{lang}.svg")';
    }
    return this.utils.stringFormat(format, params)
  },
  getAppleSvg(color) {
    const format = 'url("svg/apple-pay-{color}.svg")';
    const params = {
      color: this.getButtonColor(color),
    };
    return this.utils.stringFormat(format, params)
  },
  getAppleLabel(lang) {
    return ButtonLabelMap[lang || 'en']
  },
  addFrameImage() {
    const frame =
      this.utils.querySelector('iframe', this.element) || this.utils.createElement('iframe');
    const url = 'https://pay.google.com/gp/p/generate_gpay_btn_img';
    const style = this.params.style || {};
    const lang = this.getGoogleLangSupport(this.params.data.lang, 'en');
    const query = {
      buttonColor: style.color || 'black',
      browserLocale: lang,
      buttonSizeMode: 'fill',
    };
    const src = [url, this.utils.param(query)].join('?');
    this.addAttr(frame, {
      scrolling: 'no',
      frameborder: 0,
      src: src,
    });
    this.element.appendChild(frame);
    this.element.classList.remove('short', 'long');
  },
  styleButton() {
    let element = this.element;
    let params = this.params;
    let method = params.method;
    let style = params.style || {};
    let lang = params.data.lang || 'en';
    let css = params.css || {};
    element.setAttribute('class', '');
    element.classList.add('button', 'pending');
    if (method === 'card') method = 'google';
    if (method) {
      element.classList.add(method);
    }
    {
      element.classList.add(lang);
    }
    if (style.type) {
      element.classList.add(style.type);
    }
    if (style.mode) {
      element.classList.add(style.mode);
    }
    if (style.color) {
      element.classList.add(style.color);
    }
    if (method === 'google') {
      if (style.type === 'short') {
        style.mode = 'plain';
      }
      if (style.mode === 'default') {
        this.addFrameImage();
      } else {
        css.image = this.getGoogleSvg(style.color, lang, style.mode);
      }
    }
    if (method === 'apple') {
      css.image = this.getAppleSvg(style.color);
      css.label = this.getAppleLabel(lang, style.mode);
    }
    {
      this.utils.forEach(css, function (value, name) {
        element.style.setProperty(['--', name].join(''), value);
      });
    }
  },
  initEvents() {
    this.payment.on(
      'details',
      this.proxy(function (cx, data) {
        this.connector.send('details', data);
        this.connector.send('complete', data);
      })
    );
    this.payment.on(
      'reload',
      this.proxy(function (cx, data) {
        this.connector.send('reload', data);
      })
    );
    this.payment.on(
      'error',
      this.proxy(function (cx, data) {
        this.connector.send('error', data);
      })
    );
    this.connector.on(
      'options',
      this.proxy(function (cx, data) {
        this.extendParams(data);
        this.styleButton();
      })
    );
    this.connector.on(
      'pay',
      this.proxy(function () {
        if (!this.element.classList.contains('pending')) {
          if (this.params.method === 'apple') {
            const payload = this.payment.payload || {};
            const apple = payload.provider['apple'] || {};
            this.connector.send('pay', {
              payment_system: payload.payment_system,
              methods: apple.methods,
              details: apple.details,
              options: apple.options,
            });
          } else {
            this.payment
              .setSupported({
                fallback: false,
                provider: ['google'],
              })
              .pay('google');
          }
        }
      })
    );
    this.connector.on(
      'config',
      this.proxy(function (cx, data) {
        this.payment.setPayload((data = parseConfig(data)));
        if (data.payment_system && Object.keys(data.provider).length > 0) {
          this.element.classList.remove('pending');
          this.element.classList.add('ready');
          this.connector.send('show', {});
        } else {
          this.connector.send('hide', {});
        }
      })
    );
    this.connector.on(
      'event',
      this.proxy(function (cx, data) {
        if (data.type === 'mouseenter') {
          this.element.classList.add('hover');
        }
        if (data.type === 'mouseleave') {
          this.element.classList.remove('hover');
        }
        if (data.type === 'focusin') {
          this.element.classList.add('active');
        }
        if (data.type === 'focusout') {
          this.element.classList.remove('active');
        }
      })
    );
  },
});

const Widget = Api.extend({
  init(params) {
    this.initParams(params);
    this.initWidget();
  },
  initWidget() {
    this.initOptions(this.params.options);
    if (this.utils.isString(this.params.element)) {
      this.initElement(this.params.element);
    }
  },
  initOptions() {
    if (this.utils.isPlainObject(this.params.options)) {
      this.params.options = this.params.options || {};
    }
  },
  initElement(el) {},
  addSelectorEvent(el, ev, cb) {
    this.each(this.utils.querySelectorAll(el), function (cx, element) {
      this.addEvent(element, ev, cb);
    });
    return this
  },
  getRequestParams() {
    return {}
  },
  sendRequest(el, ev) {
    if (ev.defaultPrevented) return
    ev.preventDefault();
    this.trigger('request', this.getRequestParams(el));
    this.scope(function () {
      this.request('api.checkout.form', 'request', this.getRequestParams(el))
        .done(this.proxy('onSuccess'))
        .fail(this.proxy('onError'))
        .progress(this.proxy('onProgress'));
    });
  },
  onProgress(cx, model) {
    this.trigger('progress', model);
  },
  onSuccess(cx, model) {
    model.sendResponse();
    model.submitToMerchant();
    this.trigger('success', model);
  },
  onError(cx, model) {
    this.trigger('error', model);
  },
});

const FormData = Module.extend({
  init(form) {
    this.setFormElement(form);
  },
  setFormElement(form) {
    if (this.utils.isElement(form)) {
      this.form = form;
    }
  },
  getData(filter) {
    const params = this.deparam(this.serializeArray());
    return filter === true ? this.utils.cleanObject(params) : params
  },
  serializeArray() {
    const list = this.utils.toArray(this.form.elements);
    return this.utils.map(list, function (field) {
      if (field.disabled || field.name === '') return
      if (field.type.match('checkbox|radio') && !field.checked) return
      return {
        name: field.name,
        value: field.value,
      }
    })
  },
  serializeAndEncode() {
    return this.utils
      .map(this.serializeArray(), function (field) {
        return [field.name, encodeURIComponent(field.value)].join('=')
      })
      .join('&')
  },
  deparam(obj) {
    let prop,
      result = {};
    const breaker = /[^\[\]]+|\[\]$/g;
    const attr = function (name, value) {
      let i,
        data = result,
        last = name.pop(),
        len = name.length;
      for (i = 0; i < len; i++) {
        if (!data[name[i]]) data[name[i]] = len === i + 1 && last === '[]' ? [] : {};
        data = data[name[i]];
      }
      if (last === '[]') {
        data.push(value);
      } else {
        data[last] = value;
      }
    };
    for (prop in obj) {
      if (obj.hasOwnProperty(prop)) {
        attr(obj[prop].name.match(breaker), obj[prop].value);
      }
    }
    return result
  },
});

const WidgetForm = Widget.extend({
  initElement(el) {
    this.addSelectorEvent(el, 'submit', 'sendRequest');
  },
  getRequestParams(el) {
    return this.utils.extend({}, this.params.options, new FormData(el).getData())
  },
});

const WidgetButton = Widget.extend({
  attributes: {},
  initElement(el) {
    if (this.utils.isPlainObject(this.params.attributes)) {
      this.utils.extend(this.attributes, this.params.attributes);
    }
    this.addSelectorEvent(el, 'click', 'sendRequest');
  },
  getRequestParams(el) {
    return this.utils.extend({}, this.params.options, this.getElementData(el))
  },
  getElementData(el) {
    const result = {};
    this.utils.forEach(this.attributes, function (value, key) {
      if (el.hasAttribute(key)) {
        result[value] = el.getAttribute(key);
      }
    });
    return result
  },
});

Component.add('Api', Api);
Component.add('Connector', Connector);
Component.add('PaymentButton', PaymentButton);
Component.add('PaymentRequestApi', PaymentRequestApi);
Component.add('PaymentElement', PaymentElement);
Component.add('FormWidget', WidgetForm);
Component.add('ButtonWidget', WidgetButton);

Component.Utils = Utils;
Component.Config = Config;
Component.Api = Api;
Component.Module = Module;
Component.Connector = Connector;
Component.Response = Response;
Component.PaymentRequestApi = PaymentRequestApi;
Component.PaymentContainer = PaymentContainer;
Component.PaymentElement = PaymentElement;
Component.PaymentButton = PaymentButton;

export { Api, Config, Connector, Module, PaymentButton, PaymentContainer, PaymentElement, PaymentRequestApi, Response, Utils, Component as default };
